use strict;

# the main file for Maximum Likelihood parameter learning
# This is similar to ../minFunc/MLlearn.pl, but it doesn't use MATLAB, it uses IPOPT instead

if ($#ARGV != 0)
{
    print "Usage: $0 <options_file>\n";
    exit;
}

my $options_file = $ARGV[0];
my $output_dir;     # to figure it out depending on the input options
my $mfile;

my $line;
my $training_file;
my $use_local_machine = 1;
# determines automatically if to use the local machine or the cluster, depending on the input training file/dir
my $training_filename;
my $add_gaussian_regularizer = 0;
my $gaussian_regularizer_mean = "zero";
my $gaussian_regularizer_precision = 0;
my $gaussian_prior_file;
my $gaussian_prior_filename;
my $add_likelihood = 1;
my $weight_thermo_set;
my $thermo_XML_file;
my $thermo_XML_filename;
my $initial_params_file;
my $initial_params_filename;
my $change_variables = 0;
my $second_order_matrix_file = "none";
my $first_order_vector_file = "none";
my $which_cluster = "icics";      # can be icics or arrow
my $use_published_errors = 0;   # by default, don't use the published errors of the optical melting experiments, they give worse results
my $model_options_file = "none";    # used for the extended model
my $model_options_filename = "none";
my $num_parameters;
my $inequality_matrix_filename = "none";
my $num_ineq_constr = 0;
my $bounds = 2e19;  # default no bounds
my $bounds_in_file = "none";
my $matrix_T = "none";
my $Tconmat_file = "none";
my $include_dangles = 1;    

# make sure the ipopt bl is built
system ("cd src; make");

open (INPUT, $options_file) || die ("Cannot open $options_file");
while ($line = <INPUT>)
{
    if ($line =~ /^#/)  { next; }
    # SGE cluster: arrow
    elsif ($line =~ /SGE cluster: (\S+)/)
    {   $which_cluster = $1;  }
    #Training file: data/12rfam.txt    
    elsif ($line =~ /Training file: (\S+)/)
    {
        $training_file = $1;
        if ($training_file eq "none")
        {  $add_likelihood = 0; }
    }
    #Training filename: 12Rfam
    elsif ($line =~ /Training filename: (\S+)/)
    {   $training_filename = $1;    }
    #Add gaussian prior: 1
    elsif ($line =~ /Add gaussian regularizer: (\d)/)
    {   $add_gaussian_regularizer = $1;    }
    # Gaussian regularizer mean:
    elsif ($line =~ /Gaussian regularizer mean: (\S+)/)
    {   $gaussian_regularizer_mean = $1;    }
    #Gaussian prior variance: 1
    elsif ($line =~ /Gaussian regularizer precision: (\S+)/)
    {   $gaussian_regularizer_precision = $1;    }
    #Gaussian prior file: data/similarity.mat
    elsif ($line =~ /Gaussian prior file: (\S+)/)
    {   $gaussian_prior_file = $1;    }
    # Gaussian prior filename: SIMtetra
    elsif ($line =~ /Gaussian prior filename: (\S+)/)
    {   $gaussian_prior_filename = $1;    }
    # Weight of thermodynamic set tau: 1
    elsif ($line =~ /Weight of thermodynamic set tau: (\S+)/)
    {   $weight_thermo_set = $1;   }
    # Initial params file: data/turner_parameters_fm363_dangles0.txt
    # Initial params filename: INIT-T99-0dangles
    elsif ($line =~ /Initial params file: (\S+)/)
    {   $initial_params_file = $1;  }
    elsif ($line =~ /Initial params filename: (\S+)/)
    {   $initial_params_filename = $1;  }
    # Change variables: 1
    elsif ($line =~ /Change variables: (\d+)/)
    {   $change_variables = $1;  }
    elsif ($line =~ /Model options file: (\S+)/)
    {   $model_options_file = $1;   }
    elsif ($line =~ /Model options filename: (\S+)/)
    {   $model_options_filename = $1;   }
    # Inequality constraints matrix file: data/dangle_constraints_363T.txt
    elsif ($line =~ /Inequality constraints matrix file: (\S+)/)
    {   $inequality_matrix_filename = $1;   }    
    elsif ($line =~ /Bounds: (\S+)/)
    {
        $bounds_in_file = $1;
        if ($bounds_in_file eq "none")      {   $bounds = 2e19;     }
        else                                {   $bounds = $1;    }
    }
    elsif ($line =~ /Thermo XML file: (\S+)/)
    {   $thermo_XML_file = $1;    }
    elsif ($line =~ /Thermo XML filename: (\S+)/)
    {   $thermo_XML_filename = $1;    }    
}
close(INPUT);

# assume minFuncIC is given in the options file
# if ($minimizer eq "minFunc" && $inequality_matrix_filename ne "none")
# {
#     $minimizer = "minFuncIC";
# }
if ($inequality_matrix_filename ne "none")
{
    $num_ineq_constr = `cat $inequality_matrix_filename | wc -l`;
    chomp ($num_ineq_constr);
}

print "Number of inequality constraints: $num_ineq_constr\n";


$num_parameters = `cat $initial_params_file | wc -l`;
chomp ($num_parameters);

# figure out if we should use the local machine only or not
# if $training_file is a dir, then use SGE, otherwise use the local machine
if (-d $training_file)
{   $use_local_machine = 0; }

# create a name for the output directory
$output_dir = "results/RES_09";
if ($model_options_file ne "none")
{
    $output_dir .= "_$model_options_filename";
}
$output_dir .= "_${training_filename}_${initial_params_filename}_bnd${bounds_in_file}";
$output_dir .= "_tau-p${use_published_errors}-${weight_thermo_set}";
# if ($weight_thermo_set != 0)
# {
#     $output_dir .= "-${thermo_XML_filename}";
# }
$output_dir .= "_REG${add_gaussian_regularizer}";
if ($add_gaussian_regularizer == 1)
{
    $output_dir .= "mean${gaussian_regularizer_mean}-prec${gaussian_regularizer_precision}";
}
elsif ($add_gaussian_regularizer == 2)
{
    $output_dir .= "-${gaussian_prior_filename}";
}
$output_dir .= "_CV${change_variables}";
if (-e $output_dir)
{
    print "Directory $output_dir already exists, ABORT!\n";
    exit;
}

# type make
#system ("make");

print "Creating directory: $output_dir\n";
system ("mkdir -p $output_dir");
if (!$use_local_machine)
{    
    system ("mkdir -p $output_dir/tmp");
}
system ("cp $initial_params_file $output_dir/initial_params.txt");
# also copy the options file to this directory
system ("cp $options_file $output_dir");
# copy the executable into the new directory, otherwise it'll be hard to modify it and pretend it's still working
#system ("cp ../../MultiRNAFold/evaluate_function_smart_map  $output_dir/evaluate_function");
# copy the bl sources and this file, just to keep a record of what we've actually ran
# don't need this
#system ("cp ../../MultiRNAFold/evaluate_function_smart_map.cpp  $output_dir");
#system ("cp bl.hpp bl.cpp bl_main.cpp BLlearn_ipopt.pl  $output_dir");
#system ("cp Makefile  $output_dir/Makefile_for_bl");

my $reg_mu;
if ($gaussian_regularizer_mean eq "zero")           {   $reg_mu = 0;    }
elsif ($gaussian_regularizer_mean eq "Turner99")    {   $reg_mu = 1;    }
else    {   die "Regularizer mean must be zero or Turner99";    }

`date >> $output_dir/output_verbose.txt`;
`uname -a >> $output_dir/output_verbose.txt`;
`cat /etc/issue.net >> $output_dir/output_verbose.txt`;

if ($weight_thermo_set > 0)
{
    # obtain the first and second order matrices
    $second_order_matrix_file = "$output_dir/A.txt";
    $first_order_vector_file = "$output_dir/B.txt";
    # matrix B also contains the zero order coefficient, just so that the function value is positive

    # first make sure ../minFunc/get_matrices is built
    #system ("cd ../minFunc/; make");
    my $command = "bin/get_matrices $thermo_XML_file $model_options_file $initial_params_file $use_published_errors $second_order_matrix_file $first_order_vector_file";
    print "Executing $command\n";
    system ($command);
}

if ($change_variables)
# Compute matrix T and the initial point
{
    my $initial_transformed = "$output_dir/initial_tranformed.txt";
    
    $matrix_T = "data/matrix_T_${thermo_XML_filename}_${num_parameters}_reg-prec${gaussian_regularizer_precision}.txt";
    my $matrix_invT = "data/matrix_invT_${thermo_XML_filename}_${num_parameters}_reg-prec${gaussian_regularizer_precision}.txt";

    my $mfile = "$output_dir/change_variables.m";
    $Tconmat_file = "$output_dir/Tconmat.txt";

    # now edit the change_variables.m file file
    open (MFILE, ">$mfile") || die ("Cannot open $mfile");
    print MFILE "clear all\n";
    print MFILE <<EOF;
    if exist('$matrix_T')
        T = load('$matrix_T');
    else
        fprintf('[Loading A...');
        A = load('$second_order_matrix_file');
        fprintf(']\\n');
        fprintf('[Diagonalizing...');
        [V,D] = eig(A+$gaussian_regularizer_precision*eye(size(A)));
        clear A;
        fprintf(']\\n[Computing T...');
        T = V*inv(sqrt(D));
        clear V D ;
        save('$matrix_T','T','-ASCII') ;
        fprintf(']\\n');
    end

    if exist('$matrix_invT')
        Tinv = load('$matrix_invT');
    else
        fprintf('[Computing Tinv...');
        Tinv=inv(T);
        save('$matrix_invT','Tinv','-ASCII');
        fprintf(']\\n');
    end

    x0=load('$initial_params_file');
    fprintf('[Computing x0 transformed...');
    x0=Tinv*x0;
    fprintf(']\\n');
    clear Tinv ;
    save('$initial_transformed','x0','-ASCII');
EOF

    # now create the dangling constraints file 
    system("bin/create_dangling_constraints_ML $model_options_file > data/dangle_constraints_${num_parameters}.txt");
    
    print MFILE <<EOF;
    C = load('data/dangle_constraints_${num_parameters}.txt');
    CT = C*T;
    save ('$Tconmat_file','CT','-ASCII');

EOF

    $initial_params_file = $initial_transformed;
    close (MFILE);
    system ("matlab -nodisplay < $mfile >> $output_dir/output_verbose.txt 2>&1 ");
}

# get include_dangles
my $r = `cat $model_options_file | grep parsi_dangles`;
if ($r =~ /parsi_dangles = (\d)/)
{
    if ($1 == 1)  { $include_dangles = 0;   }  # no dangles
    else          { $include_dangles = 1;   }  # dangles is LAVISH or T99
}

print "Including dangles: $include_dangles\n";

my $ipopt_command = "./bl_main $use_local_machine $num_parameters $training_file $weight_thermo_set $reg_mu $gaussian_regularizer_precision $initial_params_file $output_dir $model_options_file $change_variables $bounds $matrix_T $second_order_matrix_file $first_order_vector_file $Tconmat_file $which_cluster $include_dangles $add_gaussian_regularizer >> $output_dir/output_verbose.txt 2>&1 &";

print "Executing $ipopt_command ...\n";
system ($ipopt_command);





#system ("nice matlab -nodisplay < $mfile > $output_dir/ml_output.txt &");

