// Copyright (C) 2005, 2006 International Business Machines and others.
// All Rights Reserved.
// This code is published under the Common Public License.
//
// $Id: hs071_nlp.cpp 949 2007-03-27 00:41:26Z andreasw $
//
// Authors:  Carl Laird, Andreas Waechter     IBM    2005-08-16

#include "bl.hpp"
#include <string.h>
#include "math.h"

using namespace Ipopt;

BL::BL ()
{
}

void read_matrix_from_file (int num_variables, char *filename, double **&matrix)
// allocate and fill up matrix, from filename
{    
    FILE *file;
    if ((matrix = new double* [num_variables]) == NULL)
    {
        printf ("Cannot allocate matrix\n");
        exit(1);
    }
    for (int i=0; i < num_variables; i++)   matrix[i] = new double[num_variables];
    
    if ((file = fopen (filename, "r")) == NULL)
    {
        printf ("Cannot open file %s\n", filename);
        exit (0);
    }           
    for (int i = 0; i < num_variables; i++)
    {
        for (int j = 0; j < num_variables; j++)
            fscanf (file, "%lf", &matrix[i][j]);
    }
    fclose (file);
}

void read_vector_from_file (int num_variables, char *filename, double *&vector)
// allocate and fill up vector, from filename
{    
    vector = new double [num_variables];
    FILE *file;
    if ((file = fopen (filename, "r")) == NULL)
    {
        printf ("Cannot open file %s\n", filename);
        exit (0);
    }           
    for (int i = 0; i < num_variables; i++)
    {
        fscanf (file, "%lf", &vector[i]);
    }
    fclose (file);
}

void matrix_times_vector (int num_variables, double **matrix, const Number *x, Number *result)
// assume result is already allocated. It'll overwrite what's in it
{
    int i,j;
    
    for (i=0; i < num_variables; i++)
    {
        result[i] = 0.0;
        
        fflush (stdout);
        for (j=0; j < num_variables; j++)
        {
            result[i] += matrix[i][j]*x[j];
        }
    }
}
   
void vector_times_matrix (int num_variables, const Number *x, double **matrix, Number *result)
// assume result is already allocated. It'll overwrite what's in it
{
    int i,j;
    
    for (j=0; j < num_variables; j++)
    {
        result[j] = 0.0;
        
        //fflush (stdout);
        for (i=0; i < num_variables; i++)
        {
            result[j] += x[i]*matrix[i][j];
        }
    }
}


void print_vector_in_file (int num_variables, const Number *x, char *filename)
{
    FILE *file;
    int i;
    if ((file = fopen (filename, "w")) == NULL)
    {
        printf ("IN print_vector_in_file: Cannot open file %s\n", filename);
        exit (1);
    }
    for (i=0; i < num_variables; i++)
    {
        fprintf (file, "%.15f\n", (double)(x[i]));
    }
    fclose (file);
}


// constructor
BL::BL (int use_local_machine, int num_variables, int num_constraints, int num_jacobian, char *struct_dir, double thermo_precision, int reg_mean, double reg_precision, char *initial_params_file, char *newdir, char *model_options_file, int change_variables, double bounds, char *Tfile, char *thermoAfile, char *thermoBfile, char *Tconmat_file, char *cluster, int add_gaussian_regularizer)
{
    this->use_local_machine = use_local_machine;
    this->num_variables = num_variables;
    this->num_constraints = num_constraints;
    this->num_jacobian = num_jacobian;
    this->add_gaussian_regularizer = add_gaussian_regularizer;
    current_iteration = 0;
    this->struct_dir = struct_dir;
    this->initial_params_file = initial_params_file;
    initial_parameters = new double[num_variables];
    mu_regularizer = reg_mean;
    this->bounds = bounds;
    strcpy (this->thermoAfile, thermoAfile);
    strcpy (this->thermoBfile, thermoBfile);
    strcpy (this->cluster, cluster);
    
    //printf ("CONSTRUCTOR: Initial params are from file %s\n", initial_params_file);
    this->newdir = newdir;
    this->thermo_precision = thermo_precision;
    this->reg_precision = reg_precision;
    this->model_options_file = model_options_file;
//     if (num_variables == 315)   ignore_dangles = 1;     // actually it doesn't matter because I use the model_options_file
//     else                        ignore_dangles = 0;
    this->change_variables = change_variables;
    strcpy (this->Tfile, Tfile);
    
    myx = new Number[num_variables];
    for (int i=0; i < num_variables; i++)
    {
        myx[i] = 0.0;
    }
       
    //NOT USED ANY MORE // strcpy (thermodynamic_file, "../../RNA-thermodynamic-database/RNA-thermo-db_v1.3.xml");

    if (change_variables)
    {
        read_matrix_from_file (num_variables, Tfile, Tmatrix);
    }

    if (num_constraints > 0)
    {
        read_matrix_from_file (num_variables, Tconmat_file, Tconmat);
    }

    if (thermo_precision != 0)
    {
        read_matrix_from_file (num_variables, thermoAfile, thermoA);
        read_vector_from_file (num_variables+1, thermoBfile, thermoB);
    }

    if (add_gaussian_regularizer == 2)
    {
        read_matrix_from_file (num_variables, "data/invSimilaritySigma_7627.txt", invSimilaritySigma);
        read_vector_from_file (num_variables, "data/similarityMu_7627.txt", similarityMu);
        tmpSim = new double[num_variables];
        tmpSim_xminusmu = new double[num_variables];
    }

}



//destructor
BL::~BL()
{
    delete [] initial_parameters;
    delete [] myx;

    if (change_variables)
    {
        for (int i=0; i < num_variables; i++)
        {
            delete [] Tmatrix[i];
        }
        delete [] Tmatrix;
    }
    if (num_constraints > 0)
    {
        for (int i=0; i < num_variables; i++)
        {
            delete [] Tconmat[i];
        }
        delete [] Tconmat;
    }
    if (thermo_precision != 0)
    {
        for (int i=0; i < num_variables; i++)
        {
            delete [] thermoA[i];
        }
        delete [] thermoA;
        delete [] thermoB;
    }
    if (add_gaussian_regularizer == 2)
    {
	for (int i=0; i < num_variables; i++)
        {
            delete [] invSimilaritySigma[i];
        }
        delete [] invSimilaritySigma;
        delete [] similarityMu;
        delete [] tmpSim;
        delete [] tmpSim_xminusmu;
    }
}


// returns the size of the problem
bool BL::get_nlp_info(Index& n, Index& m, Index& nnz_jac_g,
                             Index& nnz_h_lag, IndexStyleEnum& index_style)
{
    // the number of variables
    n = num_variables;
    
    // the number of constraints
    m = num_constraints;
    
    // the number of non-zeros in the jacobian
    nnz_jac_g = num_jacobian;
    
    // I don't know how many nonzero hessian values there are, I use quasi-newton
    //nnz_h_lag = 10;
    
    // use the C style indexing (0-based)
    index_style = TNLP::C_STYLE;

    //printf ("In GET_NLP_INFO\n");
    return true;
}

// returns the variable bounds
bool BL::get_bounds_info(Index n, Number* x_l, Number* x_u,
                                Index m, Number* g_l, Number* g_u)
{
    // here, the n and m we gave IPOPT in get_nlp_info are passed back to us.
    // If desired, we could assert to make sure they are what we think they are.
    assert(n == num_variables);
    assert(m == num_constraints);

    for (Index i=0; i < n; i++)
    {            
	x_l[i] = -bounds;
	x_u[i] = bounds;
    }

    // Now the bounds on the constraints
    for (Index i=0; i < m; i++)
    {
	g_l[i] = -2e19;
	g_u[i] = 0;
    }

    // NOTE: If there are constraints and !change_variables, something more complicated has to be done, see below
    
    /*
    if (num_variables == 315)   // set bounds infinity  (or maybe less)
    {
        for (Index i=0; i < n; i++)
        {            
            x_l[i] = -bounds;
            x_u[i] = bounds;
        }
    }
    else if (num_variables == 363)
    {
        if (change_variables)   // leave all unbounded, and add constraint
        {
            for (Index i=0; i < n; i++)
            {
                x_l[i] = -bounds;
                x_u[i] = bounds;
            }
        }
        else
        {
            // the dangling ends must be bounded above by 0
            for (Index i=0; i < 260; i++)
            {
                x_l[i] = -2e19;
                x_u[i] = 2e19;
            }
            for (Index i=260; i < 308; i++)
            {
                x_l[i] = -2e19;
                x_u[i] = 0;
            }
            for (Index i=308; i < n; i++)
            {
                x_l[i] = -2e19;
                x_u[i] = 2e19;
            }
        }
        // Now the bounds on the constraints
        for (Index i=0; i < m; i++)
        {
            g_l[i] = -2e19;
            g_u[i] = 0;
        }
    }
    */

    //printf ("IN GET_BOUNDS_INFO\n");

    return true;
}

// returns the initial point for the problem
bool BL::get_starting_point(Index n, bool init_x, Number* x,
                                   bool init_z, Number* z_L, Number* z_U,
                                   Index m, bool init_lambda,
                                   Number* lambda)
{
    // Here, we assume we only have starting values for x, if you code
    // your own NLP, you can provide starting values for the dual variables
    // if you wish
    assert(n == num_variables);
    assert(init_x == true);
    assert(init_z == false);
    assert(init_lambda == false);
    
    // initialize the starting point from file initial_params_file

    FILE *file;
    char buffer[100];
    double params[8000];
    int i;
    //printf ("STARTing point file is %s\n", initial_params_file);
    if ((file = fopen (initial_params_file, "r")) == NULL)
    {
        printf ("Cannot open file %s\n", initial_params_file);
        exit (1);
    }
    i = 0;
    fgets (buffer, sizeof(buffer), file);
    while (!feof (file))
    {        
        sscanf (buffer, "%lf\n", &params[i++]);
        fgets (buffer, sizeof(buffer), file);
    }
    assert (i == num_variables);    
    
    for (i=0; i < num_variables; i++)
    {
        //initial_parameters[i] = params[i];
        x[i] = params[i];
    }
    fclose (file);

    if (mu_regularizer != 0)    // 0 mu
    {
        // load the Turner99 parameters, because the initial ones might not be the 99 parameters
        if ((file = fopen ("data/initial_parameters_363T.txt", "r")) == NULL)
        {
            printf ("Cannot open file %s\n", "data/initial_parameters_363T.txt");
            exit (1);
        }
        i = 0;
        fgets (buffer, sizeof(buffer), file);
        while (!feof (file))
        {        
            sscanf (buffer, "%lf\n", &initial_parameters[i++]);
            fgets (buffer, sizeof(buffer), file);
        }
        assert (i == num_variables);    
        
        fclose (file);

        // don't need to "untransform", because now I am loading the turner99 parameters directly
        // If I'm loading the initial parameters, those have been transformed in BLlearn_ipopt.pl, so I have to untransform them
        // save the *Untransformed* initial point to use it in the regularizer
        //matrix_times_vector (num_variables, Tmatrix, params, initial_parameters);
    }
    
    //printf ("In GET_STARTING_POINT\n");
     
    return true;
}

// returns the value of the objective function
bool BL::eval_f(Index n, const Number* x, bool new_x, Number& obj_value)
{
    // ignore new_x
    assert(n == num_variables);
    // call an sge script that evaluates the function value at x
    // First write x in a file
    FILE *file;
    char param_file[1000];
    char fg_file[1000];
    char buffer[100];
    char command[2000];
    int i, j;
    obj_value = 0;

    if (change_variables)
    {
        // do variable transformation: myx = T*x
        matrix_times_vector (num_variables, Tmatrix, x, myx);  // myx is theta, x is psi
    }
    else
    {
        for (i=0; i < num_variables; i++)   myx[i] = x[i];
    }        

    if (strcmp (struct_dir, "none") != 0)
    {
        // write myx in a file param_file               
        sprintf (param_file, "%s/params%d.txt", newdir, current_iteration);
        print_vector_in_file (num_variables, myx, param_file);
    //     print_vector_in_file (num_variables, x, param_file);
    
        sprintf (fg_file, "%s/f%d.txt", newdir, current_iteration);
        
        if (use_local_machine)
        {
            // I'm adding the thermo precision here, not inside evaluate_function!!
            //strcpy (Tfile, "data/matrix-T_prec-1.txt");
            // one file, so execute it locally
            //evaluate_function <which_param> {f|g|b} <training_set> <parameter_file> <thermodynamic_file> <simple_model:1|0> <tau> <ignore_dangles> <numerical:{1|0}> <model_options_file> <T-matrix-file>
            //sprintf(command, "%s/evaluate_function -1 f %s %s %s 0 %lf %d 0 %s data/matrix-T_prec-1.txt > %s",
            //    newdir, struct_dir, param_file, thermodynamic_file, thermo_precision, ignore_dangles, model_options_file, fg_file);
            sprintf(command, "%s/evaluate_function -1 f %s %s none 0 0 %d 0 %s > %s",
                newdir, struct_dir, param_file, ignore_dangles, model_options_file, fg_file);
            // ignore_dangles doesn't matter because I have model_options_file
                
        }
        else
        {
            // CAUTION!  DO NOT USE themro_precision and thermodynamic_file here, add it after, separately, just like in minFunc/ml_loglikelihood.m.
            // Otherwise it is added for every array job used on the cluster
            // <parameter_file> <input_dir_manyfiles> <output_dir> <output_result_file> <weightThermoSet> <thermoXMLFile> <numerical_gradient> <which_param> <which_cluster> <model_options_file> {f|g|b} <ignore_dangles> <matrix_T_file>        
            sprintf(command, "perl dangle_compute_f_and_gradient.pl %s %s %s %s 0 none 0 -1 %s %s f %d",
                param_file, struct_dir, newdir, fg_file, cluster, model_options_file, ignore_dangles);
            // ignore_dangles doesn't matter because I have model_options_file
        }
                
        //printf ("IN EVAL_F: Iteration %d ---- Executing %s\n", current_iteration, command);
        system (command);
        
        // now read the function value from the output file
        if ((file = fopen (fg_file, "r")) == NULL)
        {
            printf ("IN EVAL_F: Cannot open file %s\n", fg_file);
            exit (1);
        }
        fgets (buffer, sizeof(buffer), file);
        sscanf (buffer, "%lf\n", &obj_value);
        fclose (file);
    }

    if (thermo_precision != 0)
    {
        // add the thermo_set
        // follow the formula from ml_loglikelihood.m (ignore factor)
        // f = f + weightThermoSet * (x'*A*x + b'*x + c);
        // first x'*A*x
        double temp;
        double thermo_contribution = 0.0;
        for (i=0; i < num_variables; i++)
        {
            temp = 0.0;
            for (j=0; j < num_variables; j++)
            {
                temp += myx[j]*thermoA[j][i];
            }
            thermo_contribution += temp * myx[i];
        }
        // next b'x
        for (i=0; i < num_variables; i++)
        {
            thermo_contribution += thermoB[i]*myx[i];
        }
        // finally c
        thermo_contribution += thermoB[num_variables];
        obj_value += thermo_precision * thermo_contribution;
    }    

    if (add_gaussian_regularizer == 1)      // add regularizer, no similarity rules
    {
	if (reg_precision != 0)
	{
	    // add the reqularizer
	    // in Matlab (ml_loglikelihood.m) it was (ignore factor):
	    // f = f + factor*sum((x-regMean).^2)*gaussianRegularizerPrecision;        
	    double sumxsq = 0;
	    for (i=0; i < num_variables; i++)
	    {
		if (mu_regularizer == 0)    // 0 mu
		{
		    sumxsq += myx[i]*myx[i];
    //                 sumxsq += x[i]*x[i];
		}
		else        // mean is initial
		{
		    sumxsq += (myx[i]-initial_parameters[i])*(myx[i]-initial_parameters[i]);
    //                 sumxsq += (x[i]-initial_parameters[i])*(x[i]-initial_parameters[i]);
		}
	    }
	    obj_value += reg_precision*sumxsq;
	}
    }
    else if (add_gaussian_regularizer == 2)     // add regularizer with similarity rules
    {
        // In ml_loglikelihood it's like this: 
        //  f = f + 1/2*(x-mu)'*invSigma*(x-mu);
        for (i=0; i < num_variables; i++)
        {
            tmpSim_xminusmu[i] = myx[i] - similarityMu[i];
        }        

        matrix_times_vector (num_variables, invSimilaritySigma, tmpSim_xminusmu, tmpSim); 
        for (i=0; i < num_variables; i++)
	{
            obj_value += 0.5 * tmpSim_xminusmu[i] * tmpSim[i];
	}
    }

    printf ("\t\t\t\t\t\tIN EVAL_F: current iteration is %d\n", current_iteration);
    fflush (stdout);

    return true;
}

// return the gradient of the objective function grad_{x} f(x)
bool BL::eval_grad_f(Index n, const Number* x, bool new_x, Number* grad_f)
{

    // ignore new_x
    assert(n == num_variables);

    // First write x in a file
    FILE *file;
    char param_file[1000];
    char fg_file[1000];
    char buffer[100];
    char command[2000];
    int i, j;
    double *tmp_grad_f = new double[num_variables];
    for (i=0; i < num_variables; i++) tmp_grad_f[i] = 0.0;

    if (change_variables)
    {
        // do variable transformation: myx = T*x
        matrix_times_vector (num_variables, Tmatrix, x, myx);
    }
    else
    {
        for (i=0; i < num_variables; i++)   myx[i] = x[i];
    }        

    if (strcmp (struct_dir, "none") != 0)
    {
        // write myx in a file param_file               
        sprintf (param_file, "%s/params%d.txt", newdir, current_iteration);
        print_vector_in_file (num_variables, myx, param_file);
    //     print_vector_in_file (num_variables, x, param_file);
    
        sprintf (fg_file, "%s/g%d.txt", newdir, current_iteration);
    
        if (use_local_machine)
        {
            // one file, so execute it locally
            //evaluate_function <which_param> {f|g|b} <training_set> <parameter_file> <thermodynamic_file> <simple_model:1|0> <tau> <ignore_dangles> <numerical:{1|0}> <model_options_file> <T-matrix-file>
            //sprintf(command, "%s/evaluate_function -1 g %s %s %s 0 %lf %d 0 %s data/matrix-T_prec-1.txt > %s",
            //    newdir, struct_dir, param_file, thermodynamic_file, thermo_precision, ignore_dangles, model_options_file, fg_file);
            sprintf(command, "%s/evaluate_function -1 g %s %s none 0 0 %d 0 %s > %s",
                newdir, struct_dir, param_file, ignore_dangles, model_options_file, fg_file);         
            // ignore_dangles doesn't matter because I have model_options_file   
        }
        else
        {
            // CAUTION!  DO NOT USE themro_precision and thermodynamic_file here, add it after, separately, just like in minFunc/ml_loglikelihood.m.
            // Otherwise it is added for every array job used on the cluster
            // <parameter_file> <input_dir_manyfiles> <output_dir> <output_result_file> <weightThermoSet> <thermoXMLFile> <numerical_gradient> <which_param> <which_cluster> <model_options_file> {f|g|b} <ignore_dangles> <matrix_T_file>        
            sprintf(command, "perl dangle_compute_f_and_gradient.pl %s %s %s %s 0 none 0 -1 %s %s g %d",
                param_file, struct_dir, newdir, fg_file, cluster, model_options_file, ignore_dangles);
            // ignore_dangles doesn't matter because I have model_options_file
        }
        
        //printf ("IN EVAL_GRAD_F: Iteration %d ---- Executing %s\n", current_iteration, command);
        system (command);
    
        // read the gradient value
        if ((file = fopen (fg_file, "r")) == NULL)
        {
            printf ("IN EVAL_GRAD_F: Cannot open file %s\n", fg_file);
            exit (1);
        }
        i = 0;
        fgets (buffer, sizeof(buffer), file);
        while (!feof (file))
        {        
            sscanf (buffer, "%lf\n", &tmp_grad_f[i++]);
            fgets (buffer, sizeof(buffer), file);
        }
        assert (i == num_variables);
        fclose (file);
    }        

    if (thermo_precision != 0)
    {
        // add the thermo_set
        // follow the formula from ml_loglikelihood.m (ignore factor)
        // g = g + weightThermoSet * (2*A*x + b);
        for (i=0; i < num_variables; i++)
        {
            for (j=0; j < num_variables; j++)
            {
                tmp_grad_f[i] += 2 * thermo_precision * thermoA[i][j] * myx[j];  // THIS WAS x[j], I THINK IT SHOULD BE myx[j]
            }
            tmp_grad_f[i] += thermo_precision*thermoB[i];
        }
    }

    if (add_gaussian_regularizer == 1)
    {
	if (reg_precision != 0)
	{
	    // add the regularizer
	    // In Matlab is was:
	    // g = g + factor*(x-regMean)*2*gaussianRegularizerPrecision;
	    
	    for (i=0; i < num_variables; i++)
	    {
		if (mu_regularizer == 0)    // mean is 0
		{
		    tmp_grad_f[i] += myx[i]*2*reg_precision;      // CHANGED x to myx
		}
		else        // mean is initial_parameters
		{
		    tmp_grad_f[i] += (myx[i]-initial_parameters[i])*2*reg_precision;   // CHANGED x to myx
		}
	    }
	}
    }
    else if (add_gaussian_regularizer == 2)
    {
        // from ml_loglikelihood:
	//     newterm = (x-mu)'*invSigma;
        //      g = g + newterm';
        for (i=0; i < num_variables; i++)
        {
            tmpSim_xminusmu[i] = myx[i] - similarityMu[i];
        }        

        vector_times_matrix (num_variables, tmpSim_xminusmu, invSimilaritySigma, tmpSim); 
        for (i=0; i < num_variables; i++)
	{
            tmp_grad_f[i] += tmpSim[i];
	}

    }

    if (change_variables)
    {
        // grad_f = T'grad_f
        for (i=0; i < num_variables; i++)
        {
            grad_f[i] = 0.0;
            for (j=0; j < num_variables; j++)
            {
                grad_f[i] += Tmatrix[j][i] * tmp_grad_f[j];
            }
        }
    }
    else
    {
        for (i=0; i < num_variables; i++)  grad_f[i] = tmp_grad_f[i]; 
    }

    // compute Linf norm of gradient
    double Linf = 0;
    double tmp_Linf = 0;
//  The following works only for a model with 363 features. I'm removing it for now, since I don't really need it.
//     for (i=0; i < num_variables; i++)
//     {
//         if (fabs (grad_f[i]) > Linf)    Linf = fabs (grad_f[i]);
//         if (i >= 260 && i <= 307)     continue;
//         if (fabs (tmp_grad_f[i]) > tmp_Linf)    tmp_Linf = fabs (tmp_grad_f[i]);
//     }
// 
//     printf ("\t\t\t\t\t\tIN EVAL_GRAD_F: current iteration is %d, tmp_Linf is %e\n", current_iteration, tmp_Linf);
//     fflush (stdout);
    current_iteration++;
    sprintf (param_file, "%s/g_total_%d.txt", newdir, current_iteration);
    print_vector_in_file (num_variables, tmp_grad_f, param_file);
    
    delete [] tmp_grad_f;
        
    return true;
}

// return the value of the constraints: g(x)
bool BL::eval_g(Index n, const Number* x, bool new_x, Index m, Number* g)
{
    // ignore new_x
    assert(n == num_variables);
    assert(m == num_constraints);    
    if (num_constraints > 0)    // these are the dangling constraints
    {
        if (change_variables)
        {
            // CAUTION
            // the constraints are Tconmat*x <= 0
            for (int i=0; i < num_constraints; i++)
            {
                g[i] = 0;
                for (int j=0; j < num_variables; j++)
                {
                    g[i] += Tconmat[i][j]*x[j];
                }
            }
        }
        else
        {
            // the following is valid only for a model with 363 features
            if (num_variables == 363)
            {
                g[0] = x[260] - x[284];
                g[1] = x[260] - x[288];
                g[2] = x[260] - x[292];
                g[3] = x[260] - x[296];
                g[4] = x[260] - x[300];
                g[5] = x[260] - x[304];
                g[6] = x[264] - x[284];
                g[7] = x[264] - x[288];
                g[8] = x[264] - x[292];
                g[9] = x[264] - x[296];
                g[10] = x[264] - x[300];
                g[11] = x[264] - x[304];
                g[12] = x[268] - x[284];
                g[13] = x[268] - x[288];
                g[14] = x[268] - x[292];
                g[15] = x[268] - x[296];
                g[16] = x[268] - x[300];
                g[17] = x[268] - x[304];
                g[18] = x[272] - x[284];
                g[19] = x[272] - x[288];
                g[20] = x[272] - x[292];
                g[21] = x[272] - x[296];
                g[22] = x[272] - x[300];
                g[23] = x[272] - x[304];
                g[24] = x[276] - x[284];
                g[25] = x[276] - x[288];
                g[26] = x[276] - x[292];
                g[27] = x[276] - x[296];
                g[28] = x[276] - x[300];
                g[29] = x[276] - x[304];
                g[30] = x[280] - x[284];
                g[31] = x[280] - x[288];
                g[32] = x[280] - x[292];
                g[33] = x[280] - x[296];
                g[34] = x[280] - x[300];
                g[35] = x[280] - x[304];
                g[36] = x[261] - x[285];
                g[37] = x[261] - x[289];
                g[38] = x[261] - x[293];
                g[39] = x[261] - x[297];
                g[40] = x[261] - x[301];
                g[41] = x[261] - x[305];
                g[42] = x[265] - x[285];
                g[43] = x[265] - x[289];
                g[44] = x[265] - x[293];
                g[45] = x[265] - x[297];
                g[46] = x[265] - x[301];
                g[47] = x[265] - x[305];
                g[48] = x[269] - x[285];
                g[49] = x[269] - x[289];
                g[50] = x[269] - x[293];
                g[51] = x[269] - x[297];
                g[52] = x[269] - x[301];
                g[53] = x[269] - x[305];
                g[54] = x[273] - x[285];
                g[55] = x[273] - x[289];
                g[56] = x[273] - x[293];
                g[57] = x[273] - x[297];
                g[58] = x[273] - x[301];
                g[59] = x[273] - x[305];
                g[60] = x[277] - x[285];
                g[61] = x[277] - x[289];
                g[62] = x[277] - x[293];
                g[63] = x[277] - x[297];
                g[64] = x[277] - x[301];
                g[65] = x[277] - x[305];
                g[66] = x[281] - x[285];
                g[67] = x[281] - x[289];
                g[68] = x[281] - x[293];
                g[69] = x[281] - x[297];
                g[70] = x[281] - x[301];
                g[71] = x[281] - x[305];
                g[72] = x[262] - x[286];
                g[73] = x[262] - x[290];
                g[74] = x[262] - x[294];
                g[75] = x[262] - x[298];
                g[76] = x[262] - x[302];
                g[77] = x[262] - x[306];
                g[78] = x[266] - x[286];
                g[79] = x[266] - x[290];
                g[80] = x[266] - x[294];
                g[81] = x[266] - x[298];
                g[82] = x[266] - x[302];
                g[83] = x[266] - x[306];
                g[84] = x[270] - x[286];
                g[85] = x[270] - x[290];
                g[86] = x[270] - x[294];
                g[87] = x[270] - x[298];
                g[88] = x[270] - x[302];
                g[89] = x[270] - x[306];
                g[90] = x[274] - x[286];
                g[91] = x[274] - x[290];
                g[92] = x[274] - x[294];
                g[93] = x[274] - x[298];
                g[94] = x[274] - x[302];
                g[95] = x[274] - x[306];
                g[96] = x[278] - x[286];
                g[97] = x[278] - x[290];
                g[98] = x[278] - x[294];
                g[99] = x[278] - x[298];
                g[100] = x[278] - x[302];
                g[101] = x[278] - x[306];
                g[102] = x[282] - x[286];
                g[103] = x[282] - x[290];
                g[104] = x[282] - x[294];
                g[105] = x[282] - x[298];
                g[106] = x[282] - x[302];
                g[107] = x[282] - x[306];
                g[108] = x[263] - x[287];
                g[109] = x[263] - x[291];
                g[110] = x[263] - x[295];
                g[111] = x[263] - x[299];
                g[112] = x[263] - x[303];
                g[113] = x[263] - x[307];
                g[114] = x[267] - x[287];
                g[115] = x[267] - x[291];
                g[116] = x[267] - x[295];
                g[117] = x[267] - x[299];
                g[118] = x[267] - x[303];
                g[119] = x[267] - x[307];
                g[120] = x[271] - x[287];
                g[121] = x[271] - x[291];
                g[122] = x[271] - x[295];
                g[123] = x[271] - x[299];
                g[124] = x[271] - x[303];
                g[125] = x[271] - x[307];
                g[126] = x[275] - x[287];
                g[127] = x[275] - x[291];
                g[128] = x[275] - x[295];
                g[129] = x[275] - x[299];
                g[130] = x[275] - x[303];
                g[131] = x[275] - x[307];
                g[132] = x[279] - x[287];
                g[133] = x[279] - x[291];
                g[134] = x[279] - x[295];
                g[135] = x[279] - x[299];
                g[136] = x[279] - x[303];
                g[137] = x[279] - x[307];
                g[138] = x[283] - x[287];
                g[139] = x[283] - x[291];
                g[140] = x[283] - x[295];
                g[141] = x[283] - x[299];
                g[142] = x[283] - x[303];
                g[143] = x[283] - x[307];
            }
        }
    }
    return true;
}

// return either the structure or values of the jacobian, depending on whether values is NULL or not
bool BL::eval_jac_g(Index n, const Number* x, bool new_x,
                           Index m, Index nele_jac, Index* iRow, Index *jCol,
                           Number* values)
{
    // ignore new_x
    assert (n == num_variables);
    assert (m == num_constraints);
    assert (nele_jac == num_jacobian);
    
    if (values == NULL)
    {
        if (num_constraints > 0)
        {
            if (change_variables)
            {
                // it is a dense jacobian
                int num = 0;
                // the constraints are Tconmat*x <= 0
                for (int i=0; i < num_constraints; i++)
                {
                    for (int j=0; j < num_variables; j++)
                    {
                        iRow[num] = i;      jCol[num] = j;      num++;
                    }
                }
            }
            else
            {
                if (num_variables == 363)
                {
                    // return the structure of the jacobian
                    for (Index i=0; i < 144; i++)
                    {
                        iRow[i] = i;
                        iRow[i+144] = i;
                    }
    
                    jCol[0] = 260;
                    jCol[1] = 260;
                    jCol[2] = 260;
                    jCol[3] = 260;
                    jCol[4] = 260;
                    jCol[5] = 260;
                    jCol[6] = 264;
                    jCol[7] = 264;
                    jCol[8] = 264;
                    jCol[9] = 264;
                    jCol[10] = 264;
                    jCol[11] = 264;
                    jCol[12] = 268;
                    jCol[13] = 268;
                    jCol[14] = 268;
                    jCol[15] = 268;
                    jCol[16] = 268;
                    jCol[17] = 268;
                    jCol[18] = 272;
                    jCol[19] = 272;
                    jCol[20] = 272;
                    jCol[21] = 272;
                    jCol[22] = 272;
                    jCol[23] = 272;
                    jCol[24] = 276;
                    jCol[25] = 276;
                    jCol[26] = 276;
                    jCol[27] = 276;
                    jCol[28] = 276;
                    jCol[29] = 276;
                    jCol[30] = 280;
                    jCol[31] = 280;
                    jCol[32] = 280;
                    jCol[33] = 280;
                    jCol[34] = 280;
                    jCol[35] = 280;
                    jCol[36] = 261;
                    jCol[37] = 261;
                    jCol[38] = 261;
                    jCol[39] = 261;
                    jCol[40] = 261;
                    jCol[41] = 261;
                    jCol[42] = 265;
                    jCol[43] = 265;
                    jCol[44] = 265;
                    jCol[45] = 265;
                    jCol[46] = 265;
                    jCol[47] = 265;
                    jCol[48] = 269;
                    jCol[49] = 269;
                    jCol[50] = 269;
                    jCol[51] = 269;
                    jCol[52] = 269;
                    jCol[53] = 269;
                    jCol[54] = 273;
                    jCol[55] = 273;
                    jCol[56] = 273;
                    jCol[57] = 273;
                    jCol[58] = 273;
                    jCol[59] = 273;
                    jCol[60] = 277;
                    jCol[61] = 277;
                    jCol[62] = 277;
                    jCol[63] = 277;
                    jCol[64] = 277;
                    jCol[65] = 277;
                    jCol[66] = 281;
                    jCol[67] = 281;
                    jCol[68] = 281;
                    jCol[69] = 281;
                    jCol[70] = 281;
                    jCol[71] = 281;
                    jCol[72] = 262;
                    jCol[73] = 262;
                    jCol[74] = 262;
                    jCol[75] = 262;
                    jCol[76] = 262;
                    jCol[77] = 262;
                    jCol[78] = 266;
                    jCol[79] = 266;
                    jCol[80] = 266;
                    jCol[81] = 266;
                    jCol[82] = 266;
                    jCol[83] = 266;
                    jCol[84] = 270;
                    jCol[85] = 270;
                    jCol[86] = 270;
                    jCol[87] = 270;
                    jCol[88] = 270;
                    jCol[89] = 270;
                    jCol[90] = 274;
                    jCol[91] = 274;
                    jCol[92] = 274;
                    jCol[93] = 274;
                    jCol[94] = 274;
                    jCol[95] = 274;
                    jCol[96] = 278;
                    jCol[97] = 278;
                    jCol[98] = 278;
                    jCol[99] = 278;
                    jCol[100] = 278;
                    jCol[101] = 278;
                    jCol[102] = 282;
                    jCol[103] = 282;
                    jCol[104] = 282;
                    jCol[105] = 282;
                    jCol[106] = 282;
                    jCol[107] = 282;
                    jCol[108] = 263;
                    jCol[109] = 263;
                    jCol[110] = 263;
                    jCol[111] = 263;
                    jCol[112] = 263;
                    jCol[113] = 263;
                    jCol[114] = 267;
                    jCol[115] = 267;
                    jCol[116] = 267;
                    jCol[117] = 267;
                    jCol[118] = 267;
                    jCol[119] = 267;
                    jCol[120] = 271;
                    jCol[121] = 271;
                    jCol[122] = 271;
                    jCol[123] = 271;
                    jCol[124] = 271;
                    jCol[125] = 271;
                    jCol[126] = 275;
                    jCol[127] = 275;
                    jCol[128] = 275;
                    jCol[129] = 275;
                    jCol[130] = 275;
                    jCol[131] = 275;
                    jCol[132] = 279;
                    jCol[133] = 279;
                    jCol[134] = 279;
                    jCol[135] = 279;
                    jCol[136] = 279;
                    jCol[137] = 279;
                    jCol[138] = 283;
                    jCol[139] = 283;
                    jCol[140] = 283;
                    jCol[141] = 283;
                    jCol[142] = 283;
                    jCol[143] = 283;
    
                    jCol[144+0] = 284;
                    jCol[144+1] = 288;
                    jCol[144+2] = 292;
                    jCol[144+3] = 296;
                    jCol[144+4] = 300;
                    jCol[144+5] = 304;
                    jCol[144+6] = 284;
                    jCol[144+7] = 288;
                    jCol[144+8] = 292;
                    jCol[144+9] = 296;
                    jCol[144+10] = 300;
                    jCol[144+11] = 304;
                    jCol[144+12] = 284;
                    jCol[144+13] = 288;
                    jCol[144+14] = 292;
                    jCol[144+15] = 296;
                    jCol[144+16] = 300;
                    jCol[144+17] = 304;
                    jCol[144+18] = 284;
                    jCol[144+19] = 288;
                    jCol[144+20] = 292;
                    jCol[144+21] = 296;
                    jCol[144+22] = 300;
                    jCol[144+23] = 304;
                    jCol[144+24] = 284;
                    jCol[144+25] = 288;
                    jCol[144+26] = 292;
                    jCol[144+27] = 296;
                    jCol[144+28] = 300;
                    jCol[144+29] = 304;
                    jCol[144+30] = 284;
                    jCol[144+31] = 288;
                    jCol[144+32] = 292;
                    jCol[144+33] = 296;
                    jCol[144+34] = 300;
                    jCol[144+35] = 304;
                    jCol[144+36] = 285;
                    jCol[144+37] = 289;
                    jCol[144+38] = 293;
                    jCol[144+39] = 297;
                    jCol[144+40] = 301;
                    jCol[144+41] = 305;
                    jCol[144+42] = 285;
                    jCol[144+43] = 289;
                    jCol[144+44] = 293;
                    jCol[144+45] = 297;
                    jCol[144+46] = 301;
                    jCol[144+47] = 305;
                    jCol[144+48] = 285;
                    jCol[144+49] = 289;
                    jCol[144+50] = 293;
                    jCol[144+51] = 297;
                    jCol[144+52] = 301;
                    jCol[144+53] = 305;
                    jCol[144+54] = 285;
                    jCol[144+55] = 289;
                    jCol[144+56] = 293;
                    jCol[144+57] = 297;
                    jCol[144+58] = 301;
                    jCol[144+59] = 305;
                    jCol[144+60] = 285;
                    jCol[144+61] = 289;
                    jCol[144+62] = 293;
                    jCol[144+63] = 297;
                    jCol[144+64] = 301;
                    jCol[144+65] = 305;
                    jCol[144+66] = 285;
                    jCol[144+67] = 289;
                    jCol[144+68] = 293;
                    jCol[144+69] = 297;
                    jCol[144+70] = 301;
                    jCol[144+71] = 305;
                    jCol[144+72] = 286;
                    jCol[144+73] = 290;
                    jCol[144+74] = 294;
                    jCol[144+75] = 298;
                    jCol[144+76] = 302;
                    jCol[144+77] = 306;
                    jCol[144+78] = 286;
                    jCol[144+79] = 290;
                    jCol[144+80] = 294;
                    jCol[144+81] = 298;
                    jCol[144+82] = 302;
                    jCol[144+83] = 306;
                    jCol[144+84] = 286;
                    jCol[144+85] = 290;
                    jCol[144+86] = 294;
                    jCol[144+87] = 298;
                    jCol[144+88] = 302;
                    jCol[144+89] = 306;
                    jCol[144+90] = 286;
                    jCol[144+91] = 290;
                    jCol[144+92] = 294;
                    jCol[144+93] = 298;
                    jCol[144+94] = 302;
                    jCol[144+95] = 306;
                    jCol[144+96] = 286;
                    jCol[144+97] = 290;
                    jCol[144+98] = 294;
                    jCol[144+99] = 298;
                    jCol[144+100] = 302;
                    jCol[144+101] = 306;
                    jCol[144+102] = 286;
                    jCol[144+103] = 290;
                    jCol[144+104] = 294;
                    jCol[144+105] = 298;
                    jCol[144+106] = 302;
                    jCol[144+107] = 306;
                    jCol[144+108] = 287;
                    jCol[144+109] = 291;
                    jCol[144+110] = 295;
                    jCol[144+111] = 299;
                    jCol[144+112] = 303;
                    jCol[144+113] = 307;
                    jCol[144+114] = 287;
                    jCol[144+115] = 291;
                    jCol[144+116] = 295;
                    jCol[144+117] = 299;
                    jCol[144+118] = 303;
                    jCol[144+119] = 307;
                    jCol[144+120] = 287;
                    jCol[144+121] = 291;
                    jCol[144+122] = 295;
                    jCol[144+123] = 299;
                    jCol[144+124] = 303;
                    jCol[144+125] = 307;
                    jCol[144+126] = 287;
                    jCol[144+127] = 291;
                    jCol[144+128] = 295;
                    jCol[144+129] = 299;
                    jCol[144+130] = 303;
                    jCol[144+131] = 307;
                    jCol[144+132] = 287;
                    jCol[144+133] = 291;
                    jCol[144+134] = 295;
                    jCol[144+135] = 299;
                    jCol[144+136] = 303;
                    jCol[144+137] = 307;
                    jCol[144+138] = 287;
                    jCol[144+139] = 291;
                    jCol[144+140] = 295;
                    jCol[144+141] = 299;
                    jCol[144+142] = 303;
                    jCol[144+143] = 307;
                }
            }
        }
    }
    else
    {
        if (num_constraints > 0)
        {
            if (change_variables)
            {
                // it is a dense jacobian
                int num = 0;
                // the constraints are Tconmat*x <= 0
                for (int i=0; i < num_constraints; i++)
                {
                    for (int j=0; j < num_variables; j++)
                    {
                        values[num] = Tconmat[i][j];      num++;
                    }
                }
            }
            else
            {
                // return the values of the jacobian of the constraints
                for (Index i=0; i < 144; i++)
                {
                    values[i] = 1;
                    values[i+144] = -1;
                }
            }
        }
    }
    
    return true;
}

//return either the structure or values of the hessian, depending on whether values is NULL or not
bool BL::eval_h(Index n, const Number* x, bool new_x,
                       Number obj_factor, Index m, const Number* lambda,
                       bool new_lambda, Index nele_hess, Index* iRow,
                       Index* jCol, Number* values)
{
    return false;     // if it's Quasi-Newton, just return false
}

void BL::finalize_solution(SolverReturn status,
                                  Index n, const Number* x, const Number* z_L, const Number* z_U,
                                  Index m, const Number* g, const Number* lambda,
                                  Number obj_value,
				  const IpoptData* ip_data,
				  IpoptCalculatedQuantities* ip_cq)
{
    // here is where we would store the solution to variables, or write to a file, etc
    // so we could use the solution.
    
    // For this example, we write the solution to the console

    // Let's write the solution to the file newdir/final_params.txt
    // First, we have to do the variable transformation back

    // myx = T*x, where T comes from the file data/matrix-T_prec-1.txt

    if (change_variables)
    {
        // do variable transformation: myx = T*x
        matrix_times_vector (num_variables, Tmatrix, x, myx);
    }
    else
    {
        for (int i=0; i < num_variables; i++)   myx[i] = x[i];
    }        
           
    char final[1000];
    sprintf (final, "%s/final_params.txt", newdir);
    print_vector_in_file (num_variables, myx, final);
//     print_vector_in_file (num_variables, x, final);

    // what follows is default output
    printf("\n\nSolution of the primal variables, x\n");
    for (Index i=0; i<n; i++) {
        printf("x[%d] = %e\n", i, x[i]);
    }
    
    printf("\n\nSolution of the bound multipliers, z_L and z_U\n");
    
    for (Index i=0; i<n; i++) {
        printf("z_L[%d] = %e\n", i, z_L[i]);
    }
    for (Index i=0; i<n; i++) {
        printf("z_U[%d] = %e\n", i, z_U[i]);
    }    
    
    printf("\n\nObjective value\n");
    printf("f(x*) = %e\n", obj_value);
    
    printf("\nFinal value of the constraints:\n");
    for (Index i=0; i<m ;i++) {
        printf("g(%d) = %e\n", i, g[i]);
    }
}
